import { TimedKVS } from 'timed-kvs';

/**
 * container.ts
 */
let storageID = 1;
function clearContainers() {
    storageID++;
}
function objectFactory(fn) {
    let container;
    return () => {
        return (container && container[storageID]) || ((container = {})[storageID] = fn());
    };
}

/**
 * data-storage.ts
 */
/**
 * Persistent Storage
 */
class SimpleStorage {
    constructor() {
        this.items = {};
    }
    getItem(key) {
        return this.items[key];
    }
    setItem(key, value) {
        this.items[key] = value;
    }
}
/**
 * returns a function which returns an EnvelopeKVS instance.
 */
function getStorage(expires, maxItems) {
    if (expires > 0 || maxItems > 0)
        return objectFactory(() => new TimedKVS({ expires, maxItems }));
    if (expires < 0)
        return objectFactory(() => new SimpleStorage());
}

/**
 * cache.ts
 */
function cacheFactory(options) {
    let { cache, hasher, maxItems, negativeCache, refresh, storage } = options;
    cache = +cache || 0;
    negativeCache = +negativeCache || 0;
    refresh = +refresh || 0;
    if (refresh < 0) {
        throw new Error("Invalid refresh: " + refresh);
    }
    if (hasher == null) {
        hasher = (arg) => JSON.stringify(arg);
    }
    if ("function" !== typeof hasher) {
        throw new Error("Invalid hasher: " + hasher);
    }
    return (fn) => {
        // pending items which are running and not stored in cache yet
        const pendItems = objectFactory(() => ({})); // new Object()
        // cache storage for resolved response
        const okItems = getStorage(cache, maxItems);
        // cache storage for rejected response
        const ngItems = getStorage(negativeCache, maxItems);
        return (arg) => {
            const key = hasher(arg);
            const pending = pendItems();
            return getItem().value;
            function getItem() {
                // read from cache
                const item = pending[key] || (okItems && okItems().getItem(key)) || (ngItems && ngItems().getItem(key));
                // run it and save to cache
                if (!item) {
                    return pending[key] = makeItem();
                }
                // refresh cache in background for the next coming request
                if (refresh && Date.now() > item.refresh) {
                    delete item.refresh;
                    makeItem();
                }
                return item;
            }
            function makeItem() {
                const item = {
                    value: storage ? startWithStorage() : start()
                };
                // unresolved cache could be refreshed
                if (refresh > 0)
                    item.refresh = Date.now() + refresh;
                item.value.then(() => {
                    // save it when resolved
                    if (okItems)
                        return okItems().setItem(key, item);
                }, () => {
                    // save it even when rejected
                    if (ngItems)
                        return ngItems().setItem(key, item);
                }).then(() => {
                    // the cache should be refreshed after given msec
                    if (refresh > 0)
                        item.refresh = Date.now() + refresh;
                    // the cache is ready then
                    delete pending[key];
                });
                return item;
            }
            // run the job without external storage
            function start() {
                return Promise.resolve().then(() => fn(arg));
            }
            // run the job with external storage
            function startWithStorage() {
                return Promise.resolve().then(() => storage.get(key)).then(cached => {
                    if (cached != null)
                        return cached;
                    return Promise.resolve().then(() => fn(arg)).then(result => {
                        if (result == null)
                            return result;
                        return Promise.resolve().then(() => storage.set(key, result)).then(() => result);
                    });
                });
            }
        };
    };
}

/**
 * concurrency.ts
 */
function concurrencyFactory(concurrency) {
    if (!(concurrency > 0)) {
        throw new Error("Invalid concurrency: " + concurrency);
    }
    return fn => {
        const queue = [];
        let running = 0;
        return arg => new Promise((resolve, reject) => {
            const job = () => {
                running++;
                return Promise.resolve().then(() => fn(arg)).then(resolve, reject);
            };
            queue.push(job);
            next();
        });
        function next() {
            if (running >= concurrency)
                return;
            const job = queue.shift();
            if (job)
                job().then(done, done);
        }
        function done() {
            running--;
            next();
        }
    };
}

/**
 * timeout.ts
 */
function timeoutFactory(timeout, fallback) {
    if (!(timeout > 0)) {
        throw new Error("Invalid timeout: " + timeout);
    }
    if (fallback == null) {
        fallback = () => Promise.reject(new Error("timeout: " + timeout + "ms"));
    }
    if ("function" !== typeof fallback) {
        throw new Error("Invalid timeoutFallback: " + fallback);
    }
    return fn => arg => new Promise((resolve, reject) => {
        let count = 0;
        let timer = setTimeout(() => {
            timer = null;
            if (!count++)
                return Promise.resolve().then(() => fallback(arg)).then(resolve, reject);
        }, timeout);
        Promise.resolve().then(() => fn(arg)).then(result => {
            if (timer)
                clearTimeout(timer);
            if (!count++)
                resolve(result);
        }, reason => {
            if (timer)
                clearTimeout(timer);
            if (!count++)
                reject(reason);
        });
    });
}

/**
 * async-cache-queue.ts
 */
/**
 * It returns a function to create a queue manager.
 */
function queueFactory(options) {
    const { cache, concurrency, negativeCache, storage, timeout, timeoutFallback } = options || {};
    let flow = null;
    if (timeout) {
        flow = join(flow, timeoutFactory(timeout, timeoutFallback));
    }
    if (concurrency) {
        flow = join(flow, concurrencyFactory(concurrency));
    }
    if (cache || negativeCache || storage) {
        flow = join(flow, cacheFactory(options));
    }
    if (!flow) {
        flow = fn => arg => Promise.resolve().then(() => fn(arg));
    }
    return fn => flow(fn);
}
function join(prev, filter) {
    return (prev && filter) ? (fn => filter(prev(fn))) : (prev || filter);
}
/**
 * It purges all cache storage handled in the library.
 */
function clearCache() {
    clearContainers();
}

export { clearCache, queueFactory };
