"use strict";
/**
 * cache.ts
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.cacheFactory = void 0;
const data_storage_1 = require("./data-storage");
const container_1 = require("./container");
function cacheFactory(options) {
    let { cache, hasher, maxItems, negativeCache, refresh, storage } = options;
    cache = +cache || 0;
    negativeCache = +negativeCache || 0;
    refresh = +refresh || 0;
    if (refresh < 0) {
        throw new Error("Invalid refresh: " + refresh);
    }
    if (hasher == null) {
        hasher = (arg) => JSON.stringify(arg);
    }
    if ("function" !== typeof hasher) {
        throw new Error("Invalid hasher: " + hasher);
    }
    return (fn) => {
        // pending items which are running and not stored in cache yet
        const pendItems = (0, container_1.objectFactory)(() => ({})); // new Object()
        // cache storage for resolved response
        const okItems = (0, data_storage_1.getStorage)(cache, maxItems);
        // cache storage for rejected response
        const ngItems = (0, data_storage_1.getStorage)(negativeCache, maxItems);
        return (arg) => {
            const key = hasher(arg);
            const pending = pendItems();
            return getItem().value;
            function getItem() {
                // read from cache
                const item = pending[key] || (okItems && okItems().getItem(key)) || (ngItems && ngItems().getItem(key));
                // run it and save to cache
                if (!item) {
                    return pending[key] = makeItem();
                }
                // refresh cache in background for the next coming request
                if (refresh && Date.now() > item.refresh) {
                    delete item.refresh;
                    makeItem();
                }
                return item;
            }
            function makeItem() {
                const item = {
                    value: storage ? startWithStorage() : start()
                };
                // unresolved cache could be refreshed
                if (refresh > 0)
                    item.refresh = Date.now() + refresh;
                item.value.then(() => {
                    // save it when resolved
                    if (okItems)
                        return okItems().setItem(key, item);
                }, () => {
                    // save it even when rejected
                    if (ngItems)
                        return ngItems().setItem(key, item);
                }).then(() => {
                    // the cache should be refreshed after given msec
                    if (refresh > 0)
                        item.refresh = Date.now() + refresh;
                    // the cache is ready then
                    delete pending[key];
                });
                return item;
            }
            // run the job without external storage
            function start() {
                return Promise.resolve().then(() => fn(arg));
            }
            // run the job with external storage
            function startWithStorage() {
                return Promise.resolve().then(() => storage.get(key)).then(cached => {
                    if (cached != null)
                        return cached;
                    return Promise.resolve().then(() => fn(arg)).then(result => {
                        if (result == null)
                            return result;
                        return Promise.resolve().then(() => storage.set(key, result)).then(() => result);
                    });
                });
            }
        };
    };
}
exports.cacheFactory = cacheFactory;
